@@ -1,35 +1,24 @@
 /*
-  Copyright (c) 2006 - 2007 Volker Krause <vkrause@kde.org>
-  Copyright (C) 2007 KovoKs <info@kovoks.nl>
-  Copyright (c) 2008 Thomas McGuire <thomas.mcguire@gmx.net>
+  SPDX-FileCopyrightText: 2006-2007 Volker Krause <vkrause@kde.org>
+  SPDX-FileCopyrightText: 2007 KovoKs <info@kovoks.nl>
+  SPDX-FileCopyrightText: 2008 Thomas McGuire <thomas.mcguire@gmx.net>
 
-  This library is free software; you can redistribute it and/or modify it
-  under the terms of the GNU Library General Public License as published by
-  the Free Software Foundation; either version 2 of the License, or (at your
-  option) any later version.
-
-  This library is distributed in the hope that it will be useful, but WITHOUT
-  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
-  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
-  License for more details.
-
-  You should have received a copy of the GNU Library General Public License
-  along with this library; see the file COPYING.LIB.  If not, write to the
-  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
-  02110-1301, USA.
+  SPDX-License-Identifier: LGPL-2.0-or-later
 */
 
 // Own
 #include "servertest.h"
 #include "socket.h"
 
-#include <transportbase.h>
 #include <mailtransport_defs.h>
+#include <transportbase.h>
 
 // Qt
 #include <QHash>
 #include <QHostInfo>
 #include <QProgressBar>
+#include <QRegularExpression>
+#include <QSet>
 #include <QTimer>
 
 // KDE
@@ -39,47 +28,51 @@ using namespace MailTransport;
 
 namespace MailTransport
 {
-
 class ServerTestPrivate
 {
 public:
     ServerTestPrivate(ServerTest *test);
 
-    ServerTest *const              q;
-    QString                        server;
-    QString                        fakeHostname;
-    QString                        testProtocol;
+    ServerTest *const q;
+    QString server;
+    QString fakeHostname;
+    QString testProtocol;
 
-    MailTransport::Socket         *normalSocket;
-    MailTransport::Socket         *secureSocket;
+    MailTransport::Socket *normalSocket = nullptr;
+    MailTransport::Socket *secureSocket = nullptr;
 
-    QSet< int >                    connectionResults;
-    QHash< int, QList<int> >       authenticationResults;
-    QSet< ServerTest::Capability > capabilityResults;
-    QHash< int, uint >             customPorts;
-    QTimer                        *normalSocketTimer;
-    QTimer                        *secureSocketTimer;
-    QTimer                        *progressTimer;
+    QSet<int> connectionResults;
+    QHash<int, QList<int>> authenticationResults;
+    QSet<ServerTest::Capability> capabilityResults;
+    QHash<int, uint> customPorts;
+    QTimer *normalSocketTimer = nullptr;
+    QTimer *secureSocketTimer = nullptr;
+    QTimer *progressTimer = nullptr;
 
-    QProgressBar                  *testProgress;
+    QProgressBar *testProgress = nullptr;
 
-    bool                           secureSocketFinished;
-    bool                           normalSocketFinished;
-    bool                           tlsFinished;
-    bool                           popSupportsTLS;
-    int                            normalStage;
-    int                            secureStage;
-    int                            encryptionMode;
+    bool secureSocketFinished = false;
+    bool normalSocketFinished = false;
+    bool tlsFinished = false;
+    bool popSupportsTLS;
+    int normalStage;
+    int secureStage;
+    int encryptionMode;
 
-    bool                           normalPossible;
-    bool                           securePossible;
+    bool normalPossible = true;
+    bool securePossible = true;
 
     void finalResult();
     void handleSMTPIMAPResponse(int type, const QString &text);
     void sendInitialCapabilityQuery(MailTransport::Socket *socket);
-    bool handlePopConversation(MailTransport::Socket *socket, int type, int stage,
-                               const QString &response, bool *shouldStartTLS);
-    QList< int > parseAuthenticationList(const QStringList &authentications);
+    bool handlePopConversation(MailTransport::Socket *socket, int type, int stage, const QString &response, bool *shouldStartTLS);
+    bool handleNntpConversation(MailTransport::Socket *socket, int type, int *stage, const QString &response, bool *shouldStartTLS);
+    QList<int> parseAuthenticationList(const QStringList &authentications);
+
+    [[nodiscard]] inline bool isGmail(const QString &server) const
+    {
+        return server.endsWith(QLatin1String("gmail.com")) || server.endsWith(QLatin1String("googlemail.com"));
+    }
 
     // slots
     void slotNormalPossible();
@@ -91,13 +84,10 @@ public:
     void slotReadSecure(const QString &text);
     void slotUpdateProgress();
 };
-
 }
 
 ServerTestPrivate::ServerTestPrivate(ServerTest *test)
-    : q(test), testProgress(Q_NULLPTR), secureSocketFinished(false),
-      normalSocketFinished(false), tlsFinished(false),
-      normalPossible(true), securePossible(true)
+    : q(test)
 {
 }
 
@@ -109,26 +99,31 @@ void ServerTestPrivate::finalResult()
 
     qCDebug(MAILTRANSPORT_LOG) << "Modes:" << connectionResults;
     qCDebug(MAILTRANSPORT_LOG) << "Capabilities:" << capabilityResults;
-    qCDebug(MAILTRANSPORT_LOG) << "Normal:" <<  q->normalProtocols();
-    qCDebug(MAILTRANSPORT_LOG) << "SSL:" <<  q->secureProtocols();
-    qCDebug(MAILTRANSPORT_LOG) << "TLS:" <<  q->tlsProtocols();
+    qCDebug(MAILTRANSPORT_LOG) << "Normal:" << q->normalProtocols();
+    qCDebug(MAILTRANSPORT_LOG) << "SSL:" << q->secureProtocols();
+    qCDebug(MAILTRANSPORT_LOG) << "TLS:" << q->tlsProtocols();
 
     if (testProgress) {
         testProgress->hide();
     }
     progressTimer->stop();
-    secureSocketFinished =  false;
-    normalSocketFinished =  false;
-    tlsFinished = false ;
+    secureSocketFinished = false;
+    normalSocketFinished = false;
+    tlsFinished = false;
 
-    emit q->finished(connectionResults.toList());
+    QList<int> resultsAsVector;
+    resultsAsVector.reserve(connectionResults.size());
+    for (int res : std::as_const(connectionResults)) {
+        resultsAsVector.append(res);
+    }
+
+    Q_EMIT q->finished(resultsAsVector);
 }
 
-QList< int > ServerTestPrivate::parseAuthenticationList(const QStringList &authentications)
+QList<int> ServerTestPrivate::parseAuthenticationList(const QStringList &authentications)
 {
-    QList< int > result;
-    for (QStringList::ConstIterator it = authentications.begin();
-            it != authentications.end(); ++it) {
+    QList<int> result;
+    for (QStringList::ConstIterator it = authentications.begin(); it != authentications.end(); ++it) {
         QString current = (*it).toUpper();
         if (current == QLatin1String("LOGIN")) {
             result << Transport::EnumAuthenticationType::LOGIN;
@@ -144,6 +139,10 @@ QList< int > ServerTestPrivate::parseAuthenticationList(const QStringList &authe
             result << Transport::EnumAuthenticationType::GSSAPI;
         } else if (current == QLatin1String("ANONYMOUS")) {
             result << Transport::EnumAuthenticationType::ANONYMOUS;
+        } else if (current == QLatin1String("XOAUTH2")) {
+            if (isGmail(server)) {
+                result << Transport::EnumAuthenticationType::XOAUTH2;
+            }
         }
         // APOP is handled by handlePopConversation()
     }
@@ -167,10 +166,12 @@ void ServerTestPrivate::handleSMTPIMAPResponse(int type, const QString &text)
     }
 
     QStringList protocols;
-    protocols << QStringLiteral("LOGIN") << QStringLiteral("PLAIN")
-              << QStringLiteral("CRAM-MD5") << QStringLiteral("DIGEST-MD5")
-              << QStringLiteral("NTLM") << QStringLiteral("GSSAPI")
-              << QStringLiteral("ANONYMOUS");
+    if (isGmail(server)) {
+        protocols << QStringLiteral("XOAUTH2");
+    }
+
+    protocols << QStringLiteral("LOGIN") << QStringLiteral("PLAIN") << QStringLiteral("CRAM-MD5") << QStringLiteral("DIGEST-MD5") << QStringLiteral("NTLM")
+              << QStringLiteral("GSSAPI") << QStringLiteral("ANONYMOUS");
 
     QStringList results;
     for (int i = 0; i < protocols.count(); ++i) {
@@ -182,7 +183,7 @@ void ServerTestPrivate::handleSMTPIMAPResponse(int type, const QString &text)
     authenticationResults[type] = parseAuthenticationList(results);
 
     // if we couldn't parse any authentication modes, default to clear-text
-    if (authenticationResults[type].size() == 0) {
+    if (authenticationResults[type].isEmpty()) {
         authenticationResults[type] << Transport::EnumAuthenticationType::CLEAR;
     }
 
@@ -199,9 +200,7 @@ void ServerTestPrivate::sendInitialCapabilityQuery(MailTransport::Socket *socket
 {
     if (testProtocol == IMAP_PROTOCOL) {
         socket->write(QStringLiteral("1 CAPABILITY"));
-
     } else if (testProtocol == SMTP_PROTOCOL) {
-
         // Detect the hostname which we send with the EHLO command.
         // If there is a fake one set, use that, otherwise use the
         // local host name (and make sure it contains a domain, so the
@@ -225,56 +224,48 @@ void ServerTestPrivate::sendInitialCapabilityQuery(MailTransport::Socket *socket
 
 void ServerTestPrivate::slotTlsDone()
 {
-
     // The server will not send a response after starting TLS. Therefore, we have to manually
     // call slotReadNormal(), because this is not triggered by a data received signal this time.
     slotReadNormal(QString());
 }
 
-bool ServerTestPrivate::handlePopConversation(MailTransport::Socket *socket, int type, int stage,
-        const QString &response, bool *shouldStartTLS)
+bool ServerTestPrivate::handlePopConversation(MailTransport::Socket *socket, int type, int stage, const QString &response, bool *shouldStartTLS)
 {
-    Q_ASSERT(shouldStartTLS != 0);
+    Q_ASSERT(shouldStartTLS != nullptr);
 
     // Initial Greeting
     if (stage == 0) {
-
-        //Regexp taken from POP3 ioslave
-        QString responseWithoutCRLF = response;
-        responseWithoutCRLF.chop(2);
-        QRegExp re(QStringLiteral("<[A-Za-z0-9\\.\\-_]+@[A-Za-z0-9\\.\\-_]+>$"),
-                   Qt::CaseInsensitive);
+        // Regexp taken from POP3 ioslave
+        const QString responseWithoutCRLF = response.isEmpty() ? response : response.chopped(2);
+        static const QRegularExpression re(QStringLiteral("<[A-Za-z0-9\\.\\-_]+@[A-Za-z0-9\\.\\-_]+>$"), QRegularExpression::CaseInsensitiveOption);
         if (responseWithoutCRLF.indexOf(re) != -1) {
             authenticationResults[type] << Transport::EnumAuthenticationType::APOP;
         }
 
-        //Each server is supposed to support clear text login
+        // Each server is supposed to support clear text login
         authenticationResults[type] << Transport::EnumAuthenticationType::CLEAR;
 
         // If we are in TLS stage, the server does not send the initial greeting.
         // Assume that the APOP availability is the same as with an unsecured connection.
-        if (type == Transport::EnumEncryption::TLS &&
-                authenticationResults[Transport::EnumEncryption::None].
-                contains(Transport::EnumAuthenticationType::APOP)) {
-            authenticationResults[Transport::EnumEncryption::TLS]
-                    << Transport::EnumAuthenticationType::APOP;
+        if (type == Transport::EnumEncryption::TLS
+            && authenticationResults[Transport::EnumEncryption::None].contains(Transport::EnumAuthenticationType::APOP)) {
+            authenticationResults[Transport::EnumEncryption::TLS] << Transport::EnumAuthenticationType::APOP;
         }
 
         socket->write(QStringLiteral("CAPA"));
         return true;
     }
-
     // CAPA result
     else if (stage == 1) {
-//     Example:
-//     CAPA
-//     +OK
-//     TOP
-//     USER
-//     SASL LOGIN CRAM-MD5
-//     UIDL
-//     RESP-CODES
-//     .
+        //     Example:
+        //     CAPA
+        //     +OK
+        //     TOP
+        //     USER
+        //     SASL LOGIN CRAM-MD5
+        //     UIDL
+        //     RESP-CODES
+        //     .
         if (response.contains(QLatin1String("TOP"))) {
             capabilityResults += ServerTest::Top;
         }
@@ -291,35 +282,94 @@ bool ServerTestPrivate::handlePopConversation(MailTransport::Socket *socket, int
         socket->write(QStringLiteral("AUTH"));
         return true;
     }
-
     // AUTH response
     else if (stage == 2) {
-//     Example:
-//     C: AUTH
-//     S: +OK List of supported authentication methods follows
-//     S: LOGIN
-//     S: CRAM-MD5
-//     S:.
+        //     Example:
+        //     C: AUTH
+        //     S: +OK List of supported authentication methods follows
+        //     S: LOGIN
+        //     S: CRAM-MD5
+        //     S:.
         QString formattedReply = response;
 
-        // Get rid of trailling ".CRLF"
+        // Get rid of trailing ".CRLF"
         formattedReply.chop(3);
 
         // Get rid of the first +OK line
-        formattedReply = formattedReply.right(formattedReply.size() -
-                                              formattedReply.indexOf(QLatin1Char('\n')) - 1);
-        formattedReply =
-            formattedReply.replace(QLatin1Char(' '), QLatin1Char('-')).
-            replace(QLatin1String("\r\n"), QLatin1String(" "));
+        formattedReply = formattedReply.right(formattedReply.size() - formattedReply.indexOf(QLatin1Char('\n')) - 1);
+        formattedReply = formattedReply.replace(QLatin1Char(' '), QLatin1Char('-')).replace(QLatin1String("\r\n"), QLatin1String(" "));
 
-        authenticationResults[type] +=
-            parseAuthenticationList(formattedReply.split(QLatin1Char(' ')));
+        authenticationResults[type] += parseAuthenticationList(formattedReply.split(QLatin1Char(' ')));
     }
 
     *shouldStartTLS = popSupportsTLS;
     return false;
 }
 
+bool ServerTestPrivate::handleNntpConversation(MailTransport::Socket *socket, int type, int *stage, const QString &response, bool *shouldStartTLS)
+{
+    Q_ASSERT(shouldStartTLS != nullptr);
+    Q_ASSERT(stage != nullptr);
+
+    // Initial Greeting
+    if (*stage == 0) {
+        if (response.startsWith(QLatin1String("382 "))) {
+            return true;
+        }
+        if (!response.isEmpty() && !response.startsWith(QLatin1String("200 "))) {
+            return false;
+        }
+
+        socket->write(QStringLiteral("CAPABILITIES"));
+        return true;
+    }
+    // CAPABILITIES result
+    else if (*stage == 1) {
+        // Check whether we got "500 command 'CAPABILITIES' not recognized"
+        if (response.startsWith(QLatin1String("500 "))) {
+            return false;
+        }
+
+        //     Example:
+        //     101 Capability list:
+        //     VERSION 2
+        //     IMPLEMENTATION INN 2.5.4
+        //     AUTHINFO USER SASL
+        //     HDR
+        //     LIST ACTIVE [etc]
+        //     OVER
+        //     POST
+        //     READER
+        //     SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN
+        //     STARTTLS
+        //     .
+        const QList<QStringView> lines = QStringView(response).split(QStringLiteral("\r\n"), Qt::SkipEmptyParts);
+        for (const QStringView line : lines) {
+            if (line.compare(QLatin1String("STARTTLS"), Qt::CaseInsensitive) == 0) {
+                *shouldStartTLS = true;
+            } else if (line.startsWith(QLatin1String("AUTHINFO "), Qt::CaseInsensitive)) {
+                const QList<QStringView> authinfos = QStringView(line).split(QLatin1Char(' '), Qt::SkipEmptyParts);
+                const QString s(QStringLiteral("USER"));
+                const QStringView ref(s);
+                if (authinfos.contains(ref)) {
+                    authenticationResults[type].append(Transport::EnumAuthenticationType::CLEAR); // XXX
+                }
+            } else if (line.startsWith(QLatin1String("SASL "), Qt::CaseInsensitive)) {
+                const QStringList auths = line.mid(5).toString().split(QLatin1Char(' '), Qt::SkipEmptyParts);
+                authenticationResults[type] += parseAuthenticationList(auths);
+            } else if (line == QLatin1Char('.')) {
+                return false;
+            }
+        }
+        // We have not hit the end of the capabilities list yet,
+        // so avoid the stage counter to rise without reason.
+        --(*stage);
+        return true;
+    }
+
+    return false;
+}
+
 // slotReadNormal() handles normal (no) encryption and TLS encryption.
 // At first, the communication is not encrypted, but if the server supports
 // the STARTTLS/STLS keyword, the same authentication query is done again
@@ -345,11 +395,14 @@ void ServerTestPrivate::slotReadNormal(const QString &text)
     bool shouldStartTLS = false;
     normalStage++;
 
-    // Handle the whole POP converstation separatly, it is very different from
-    // IMAP and SMTP
+    // Handle the whole POP and NNTP conversations separately, as
+    // they are very different from IMAP and SMTP
     if (testProtocol == POP_PROTOCOL) {
-        if (handlePopConversation(normalSocket, encryptionMode, normalStage, text,
-                                  &shouldStartTLS)) {
+        if (handlePopConversation(normalSocket, encryptionMode, normalStage, text, &shouldStartTLS)) {
+            return;
+        }
+    } else if (testProtocol == NNTP_PROTOCOL) {
+        if (handleNntpConversation(normalSocket, encryptionMode, &normalStage, text, &shouldStartTLS)) {
             return;
         }
     } else {
@@ -399,8 +452,12 @@ void ServerTestPrivate::slotReadSecure(const QString &text)
     secureStage++;
     if (testProtocol == POP_PROTOCOL) {
         bool dummy;
-        if (handlePopConversation(secureSocket, Transport::EnumEncryption::SSL,
-                                  secureStage, text, &dummy)) {
+        if (handlePopConversation(secureSocket, Transport::EnumEncryption::SSL, secureStage, text, &dummy)) {
+            return;
+        }
+    } else if (testProtocol == NNTP_PROTOCOL) {
+        bool dummy;
+        if (handleNntpConversation(secureSocket, Transport::EnumEncryption::SSL, &secureStage, text, &dummy)) {
             return;
         }
     } else {
@@ -416,6 +473,14 @@ void ServerTestPrivate::slotReadSecure(const QString &text)
 
 void ServerTestPrivate::slotNormalNotPossible()
 {
+    if (testProtocol == SMTP_PROTOCOL && normalSocket->port() == SMTP_PORT) {
+        // For SMTP, fallback to port 25
+        normalSocket->setPort(SMTP_OLD_PORT);
+        normalSocket->reconnect();
+        normalSocketTimer->start(10000);
+        return;
+    }
+
     normalSocketTimer->stop();
     normalPossible = false;
     normalSocketFinished = true;
@@ -447,7 +512,8 @@ void ServerTestPrivate::slotUpdateProgress()
 //---------------------- end private class -----------------------//
 
 ServerTest::ServerTest(QObject *parent)
-    : QObject(parent), d(new ServerTestPrivate(this))
+    : QObject(parent)
+    , d(new ServerTestPrivate(this))
 {
     d->normalSocketTimer = new QTimer(this);
     d->normalSocketTimer->setSingleShot(true);
@@ -461,14 +527,11 @@ ServerTest::ServerTest(QObject *parent)
     connect(d->progressTimer, SIGNAL(timeout()), SLOT(slotUpdateProgress()));
 }
 
-ServerTest::~ServerTest()
-{
-    delete d;
-}
+ServerTest::~ServerTest() = default;
 
 void ServerTest::start()
 {
-    qCDebug(MAILTRANSPORT_LOG) << d;
+    qCDebug(MAILTRANSPORT_LOG) << d.get();
 
     d->connectionResults.clear();
     d->authenticationResults.clear();
@@ -502,6 +565,9 @@ void ServerTest::start()
     } else if (d->testProtocol == POP_PROTOCOL) {
         d->normalSocket->setPort(POP_PORT);
         d->secureSocket->setPort(POPS_PORT);
+    } else if (d->testProtocol == NNTP_PROTOCOL) {
+        d->normalSocket->setPort(NNTP_PORT);
+        d->secureSocket->setPort(NNTPS_PORT);
     }
 
     if (d->customPorts.contains(Transport::EnumEncryption::None)) {
@@ -513,8 +579,7 @@ void ServerTest::start()
 
     connect(d->normalSocket, SIGNAL(connected()), SLOT(slotNormalPossible()));
     connect(d->normalSocket, SIGNAL(failed()), SLOT(slotNormalNotPossible()));
-    connect(d->normalSocket, SIGNAL(data(QString)),
-            SLOT(slotReadNormal(QString)));
+    connect(d->normalSocket, SIGNAL(data(QString)), SLOT(slotReadNormal(QString)));
     connect(d->normalSocket, SIGNAL(tlsDone()), SLOT(slotTlsDone()));
     d->normalSocket->reconnect();
     d->normalSocketTimer->start(10000);
@@ -526,8 +591,7 @@ void ServerTest::start()
         d->secureSocket->setSecure(true);
         connect(d->secureSocket, SIGNAL(connected()), SLOT(slotSslPossible()));
         connect(d->secureSocket, SIGNAL(failed()), SLOT(slotSslNotPossible()));
-        connect(d->secureSocket, SIGNAL(data(QString)),
-                SLOT(slotReadSecure(QString)));
+        connect(d->secureSocket, SIGNAL(data(QString)), SLOT(slotReadSecure(QString)));
         d->secureSocket->reconnect();
         d->secureSocketTimer->start(10000);
     } else {
@@ -550,10 +614,9 @@ void ServerTest::setServer(const QString &server)
     d->server = server;
 }
 
-void ServerTest::setPort(Transport::EnumEncryption::type encryptionMode, uint port)
+void ServerTest::setPort(Transport::EnumEncryption encryptionMode, uint port)
 {
-    Q_ASSERT(encryptionMode == Transport::EnumEncryption::None ||
-             encryptionMode == Transport::EnumEncryption::SSL);
+    Q_ASSERT(encryptionMode == Transport::EnumEncryption::None || encryptionMode == Transport::EnumEncryption::SSL);
     d->customPorts.insert(encryptionMode, port);
 }
 
@@ -578,10 +641,9 @@ QString ServerTest::server() const
     return d->server;
 }
 
-int ServerTest::port(TransportBase::EnumEncryption::type encryptionMode) const
+int ServerTest::port(TransportBase::EnumEncryption encryptionMode) const
 {
-    Q_ASSERT(encryptionMode == Transport::EnumEncryption::None ||
-             encryptionMode == Transport::EnumEncryption::SSL);
+    Q_ASSERT(encryptionMode == Transport::EnumEncryption::None || encryptionMode == Transport::EnumEncryption::SSL);
     if (d->customPorts.contains(encryptionMode)) {
         return d->customPorts.value(static_cast<int>(encryptionMode));
     } else {
@@ -594,7 +656,7 @@ QProgressBar *ServerTest::progressBar() const
     return d->testProgress;
 }
 
-QList< int > ServerTest::normalProtocols() const
+QList<int> ServerTest::normalProtocols() const
 {
     return d->authenticationResults[TransportBase::EnumEncryption::None];
 }
@@ -604,12 +666,12 @@ bool ServerTest::isNormalPossible() const
     return d->normalPossible;
 }
 
-QList< int > ServerTest::tlsProtocols() const
+QList<int> ServerTest::tlsProtocols() const
 {
     return d->authenticationResults[TransportBase::EnumEncryption::TLS];
 }
 
-QList< int > ServerTest::secureProtocols() const
+QList<int> ServerTest::secureProtocols() const
 {
     return d->authenticationResults[Transport::EnumEncryption::SSL];
 }
@@ -619,9 +681,9 @@ bool ServerTest::isSecurePossible() const
     return d->securePossible;
 }
 
-QList< ServerTest::Capability > ServerTest::capabilities() const
+QList<ServerTest::Capability> ServerTest::capabilities() const
 {
-    return d->capabilityResults.toList();
+    return d->capabilityResults.values();
 }
 
 #include "moc_servertest.cpp"
